/**
 * @file src/types/modules/modules.h
 * @brief Include file of all known modules.
 * @copyright (c) 2017 Avast Software, licensed under the MIT license
 */

#pragma once

#include "yaramod/types/modules/androguard_module.h"
#include "yaramod/types/modules/cuckoo_module.h"
#include "yaramod/types/modules/dex_module.h"
#include "yaramod/types/modules/dotnet_module.h"
#include "yaramod/types/modules/elf_module.h"
#include "yaramod/types/modules/hash_module.h"
#include "yaramod/types/modules/macho_module.h"
#include "yaramod/types/modules/magic_module.h"
#include "yaramod/types/modules/math_module.h"
#include "yaramod/types/modules/pe_module.h"
#include "yaramod/types/modules/phish_module.h"
#include "yaramod/types/modules/time_module.h"
